from typing import Callable, Dict, List, Tuple, Union

import networkx as nx
import numpy as np
from networkx.classes.reportviews import NodeDataView

from aiondao.libs.conviction import trigger_threshold
from aiondao.agents.members import (
    Participant,
    ParticipantSupport,
    Proposal,
    ProposalStatus,
)
from aiondao.agents.commons import TokenBatch


def get_edges_by_type(
    network: nx.DiGraph, edge_type_selection: str, participant_idx: int = None
):
    def filter_by_type(n1, n2):
        if (participant_idx is None or n1 == participant_idx) and network.edges[
            (n1, n2)
        ]["type"] == edge_type_selection:
            return True
        return False

    view = nx.subgraph_view(network, filter_edge=filter_by_type)
    return view.edges()


def get_edges_by_participant_and_type(
    network: nx.DiGraph, participant_idx: int, edge_type_selection: str
) -> Dict:
    edges_view = network.adj[participant_idx]
    answer = {}
    for key in edges_view:
        if edges_view[key]["type"] == edge_type_selection:
            answer[key] = edges_view[key]
    return answer


def get_proposals(network: nx.DiGraph, status: ProposalStatus = None) -> NodeDataView:
    def filter_proposal(n):
        if isinstance(network.nodes[n]["item"], Proposal):
            if status:
                return network.nodes[n]["item"].status == status
            return True
        return False

    view = nx.subgraph_view(network, filter_node=filter_proposal)
    return view.nodes(data="item")


def get_participants(network: nx.DiGraph) -> NodeDataView:
    def filter_participant(n):
        if isinstance(network.nodes[n]["item"], Participant):
            return True
        return False

    view = nx.subgraph_view(network, filter_node=filter_participant)
    return view.nodes(data="item")


def get_proposals_by_participant_and_status(
    network: nx.DiGraph, participant_idx: int, status_filter: List[ProposalStatus] = []
) -> Dict:
    edges_view = network.adj[participant_idx]
    answer = {}
    for key in edges_view:
        edge = edges_view[key]
        node = network.nodes[key]["item"]
        if edge["type"] == "support" and (
            len(status_filter) == 0 or node.status in status_filter
        ):
            answer[key] = edge
    return answer


def add_proposal(
    network: nx.DiGraph, p: Proposal, proposed_by: int, random_number_func
) -> Tuple[nx.DiGraph, int]:
    j = max(network.nodes) + 1
    network.add_node(j, item=p)
    network = setup_support_edges(network, random_number_func, j)
    # The Participant who created this Proposal must have maximum affinity and need to be marked as author
    network.add_edge(
        proposed_by,
        j,
        support=ParticipantSupport(affinity=1, tokens=0, conviction=0, is_author=True),
        type="support",
    )
    return network, j


def add_participant(
    network: nx.DiGraph, p: Participant, exponential_func, random_number_func
) -> Tuple[nx.DiGraph, int]:
    j = max(network.nodes) + 1
    network.add_node(j, item=p)
    # network = setup_influence_edges_single(network, j, exponential_func) # TODO: Disabled as these aren't being used on any model policy
    network = setup_support_edges(network, random_number_func, j)
    return network, j


def create_network(
    *, token_batches: List[TokenBatch], probability_func, random_number_func
) -> nx.DiGraph:
    """
    Creates a new DiGraph with Participants corresponding to the input
    TokenBatches.
    """
    network = nx.DiGraph()
    for i, tb in enumerate(token_batches):

        p_instance = Participant(
            holdings=tb,
            prob_fn=probability_func,
            random_number_func=random_number_func,
        )
        # Make the initial participants have sentiments between 0.5 and 1
        p_instance.sentiment = 0.5 + 0.5 * random_number_func()
        network.add_node(i, item=p_instance)
    return network


def influence(exponential_func, scale=1, sigmas=3):
    """
    Calculates the likelihood of one node having influence over another node. If
    so, it returns an influence value, else None.

    expon.rvs with the standard kwargs gives you a lot more values smaller than
    1, but quite a few outliers that could go all the way up to even 8.

    Unless your influence is 3 standard deviations above the norm (where scale
    determines the size of the standard deviation), you don't have any influence
    at all.

    This is broken out so that code that populates the graph initially and code
    that adds new Participants later on can share this code.
    """

    influence_rv = exponential_func(loc=0.0, scale=scale)
    if influence_rv > scale + sigmas * scale**2:
        return influence_rv
    return None


def setup_influence_edges_bulk(network: nx.DiGraph, exponential_func) -> nx.DiGraph:
    """
    Calculates the chances that a Participant is influential enough to have an
    'influence' edge in the network to other Participants, and creates the
    corresponding edge in the graph. If an "influence" type edge already exists,
    it will skip it.

    Takes an optional participant argument, which is the index number of the
    Participant in network.nodes. If this argument is present, it will setup the
    influence edges only for this Participant.
    """
    # Turn it into a dict to make a copy out of the View, because the View
    # changes whenever we add edges to the graph, which results in RuntimeError:
    # dictionary changed size during iteration
    participants = dict(get_participants(network))

    for i in participants:
        for other_participant in participants:
            if not (other_participant == i) and not network.has_edge(
                i, other_participant
            ):
                influence_rv = influence(exponential_func)
                if influence_rv:
                    network.add_edge(
                        i, other_participant, influence=influence_rv, type="influence"
                    )
    return network


def setup_influence_edges_single(
    network: nx.DiGraph, participant: int, exponential_func
):
    p = dict(get_participants(network))
    del p[participant]
    other_participants = p

    # If we already have Participants at index 0,1,2,3,4 and we added a
    # Participant at index 5, this creates the edges 0,5; 1,5; 2;5 etc.
    for other in other_participants:
        if not network.has_edge(other, participant):
            influence_rv = influence(exponential_func)
            if influence_rv:
                network.add_edge(
                    other, participant, influence=influence_rv, type="influence"
                )
        if not network.has_edge(participant, other):
            influence_rv = influence(exponential_func)
            if influence_rv:
                network.add_edge(
                    participant, other, influence=influence_rv, type="influence"
                )
    return network


def setup_conflict_edges(
    network: nx.DiGraph, random_number_func, proposal=None, rate=0.25
) -> nx.DiGraph:
    """
    Supporting one Proposal may mean going against another Proposal, in which
    case a Proposal-Proposal conflict edge is created. This function calculates
    the chances of that happening and the 'strength' of such a conflict.

    Takes an optional proposal argument, which is the index number of the
    Proposal in network.nodes. If this argument is present, it will setup the
    conflict edges only for this Proposal.
    """

    def loop_over_other_proposals(network, proposals, proposal, random_number_func):
        for other_proposal in proposals:
            if not (other_proposal == proposal):
                # (rate=0.25) means 25% of other Proposals are going to conflict
                # with this particular Proposal. And when they do conflict, the
                # conflict number is high (at least 1 - 0.25 = 0.75).
                conflict_rv = random_number_func()
                if conflict_rv < rate:
                    network.add_edge(proposal, other_proposal)
                    network.edges[(proposal, other_proposal)]["conflict"] = (
                        1 - conflict_rv
                    )
                    network.edges[(proposal, other_proposal)]["type"] = "conflict"
        return network

    # Turn it into a dict to make a copy out of the View, because the View
    # changes whenever we add edges to the graph, which results in RuntimeError:
    # dictionary changed size during iteration
    proposals = dict(get_proposals(network))

    # Do not use "if not proposal" - index number 0 will evaluate to False.
    if proposal is None:
        for i in proposals:
            network = loop_over_other_proposals(
                network, proposals, i, random_number_func
            )
        return network
    return loop_over_other_proposals(network, proposals, proposal, random_number_func)


def setup_support_edges(
    network: nx.DiGraph, random_number_func: Callable[..., float], idx=None
) -> nx.DiGraph:
    """
    Every Participant has a 'support' edge to every Proposal, and vice versa,
    indicating how much that Participant supports that Proposal. This function
    adds support edges between every Participant and Proposal in the network.

    Takes an optional node index. If the node is a Participant, it will setup
    support edges to other Proposal nodes and vice versa if the node is a
    Proposal.
    """

    def create_support_edge(
        n: nx.DiGraph,
        participant: Participant,
        proposal: Proposal,
        random_number_func: Callable[..., float],
    ):
        # Token Holder -> Proposal Relationship
        # Looks like Zargham skewed this distribution heavily towards
        # numbers smaller than 0.25 This is the affinity towards proposals.
        # Most Participants won't care about most proposals, but then there
        # will be a few Proposals that they really care about.
        random_value = random_number_func()
        random_affinity = 1 - 4 * (1 - random_value) * random_value
        n.add_edge(
            participant,
            proposal,
            support=ParticipantSupport(
                affinity=random_affinity, tokens=0, conviction=0
            ),
            type="support",
        )
        return n

    participants = dict(get_participants(network))

    if idx is None:
        proposals = dict(get_proposals(network))
        for prop in proposals:
            for par in participants:
                network = create_support_edge(network, par, prop, random_number_func)

    else:
        if isinstance(network.nodes[idx]["item"], Proposal):

            for par in participants:
                par: Participant
                network = create_support_edge(network, par, idx, random_number_func)
        elif isinstance(network.nodes[idx]["item"], Participant):
            proposals = dict(get_proposals(network, ProposalStatus.CANDIDATE))
            for prop in proposals:
                prop: Proposal
                network = create_support_edge(network, idx, prop, random_number_func)
    return network


def bootstrap_network(
    n_participants: List[TokenBatch],
    n_proposals: int,
    funding_pool: float,
    token_supply: float,
    max_proposal_request: float,
    probability_func: Callable[..., float],
    random_number_func: Callable[..., float],
    gamma_func: Callable[..., float],
    exponential_func: Callable[..., float],
) -> nx.DiGraph:
    """
    Convenience function that creates a network ready for simulation in
    the Python notebook in one line.
    """
    n = create_network(
        token_batches=n_participants,
        probability_func=probability_func,
        random_number_func=random_number_func,
    )

    for _ in range(n_proposals):
        idx = len(n)
        r_rv = gamma_func(3, loc=0.001, scale=10000)
        n.add_node(
            idx,
            item=Proposal(
                funds_requested=r_rv,
                trigger=trigger_threshold(
                    r_rv, funding_pool, token_supply, max_proposal_request
                ),
            ),
        )

    n = setup_support_edges(n, random_number_func)
    n = setup_conflict_edges(n, random_number_func)
    # n = setup_influence_edges_bulk(n, exponential_func)  # TODO: Disabled as these aren't being used on any model policy
    return n


def calc_total_funds_requested(network: nx.DiGraph):
    candidates = get_proposals(network, status=ProposalStatus.CANDIDATE)
    fund_requests = [j[1].funds_requested for j in candidates]
    total_funds_requested = np.sum(fund_requests)
    return total_funds_requested


def calc_median_affinity(network: nx.DiGraph):
    supporters = get_edges_by_type(network, "support")
    if len(supporters) == 0:
        raise Exception("The network has 0 support edges!")

    affinities = [network.edges[e]["support"].affinity for e in supporters]
    median_affinity = np.median(affinities)
    return median_affinity


def calc_total_conviction(network: nx.DiGraph, proposal_idx: int) -> float:
    proposal = network.nodes(data="item")[proposal_idx]
    if not isinstance(proposal, Proposal):
        raise Exception("proposal_idx must point to a node that has a Proposal")

    incoming_edges = network.in_edges(proposal_idx, data="support")
    convictions = [
        get_core_support(support.conviction)
        for _, _, support in incoming_edges
        if support
    ]

    return np.sum(convictions)


def calc_total_affinity(network: nx.DiGraph) -> float:
    view = network.edges(data="support")
    affinities = [support.affinity for _, _, support in view]
    return np.sum(affinities)


def calc_avg_sentiment(network: nx.DiGraph) -> float:
    participants = get_participants(network)
    sentiment_total = 0.0
    for _, participant in participants:
        sentiment_total += participant.sentiment

    sentiment_avg = sentiment_total / len(participants)
    return sentiment_avg


def find_in_edges_of_type_for_proposal(
    network: nx.DiGraph, proposal_idx: int, edge_type: str
) -> List[Tuple[int, int, str]]:
    ans = []
    for participant_idx, proposal_idx, t in network.in_edges(proposal_idx, data="type"):
        if t == edge_type:
            ans.append((participant_idx, proposal_idx, edge_type))

    return ans


def get_proposals_conviction_list(network: nx.DiGraph):
    """
    Convenience function. Return a list of proposals' conviction of
    a given network.
    """
    support_edges = get_edges_by_type(network, "support")
    conviction_list = []
    for i, j in support_edges:
        edge = network.edges[i, j]
        conviction = edge["support"].conviction
        conviction_list.append(conviction)
    return conviction_list


def get_core_support(expected_conviction: Union[ParticipantSupport, float]) -> float:
    """
    Convenience function. Return the value of a given proposal.
    """
    if isinstance(expected_conviction, ParticipantSupport):
        return get_core_support(expected_conviction.conviction)
    return expected_conviction



from aiondao._imports import *
from loguru import logger
import numpy as np
from inspect import getmembers
from types import FunctionType
from pydantic import BaseModel
from scipy.stats import expon, gamma

from auto_all import start_all, end_all


start_all(globals())


def new_probability_func(seed):
    random_state = np.random.RandomState(seed)

    def probability(rate):
        if rate > 1.0:
            raise Exception("Rate has a maximum value of 1.0")
        return random_state.rand() < rate

    return probability


def new_exponential_func(seed):
    random_state = np.random.RandomState(seed)

    def exponential(loc, scale):
        return expon.rvs(loc=loc, scale=scale, random_state=random_state)

    return exponential


def new_gamma_func(seed):
    random_state = np.random.RandomState(seed)

    def gamma_func(alpha, loc, scale):
        return gamma.rvs(alpha, loc=loc, scale=scale, random_state=random_state)

    return gamma_func


def new_random_number_func(seed):
    random_state = np.random.RandomState(seed)

    def random_number_func():
        return random_state.rand()

    return random_number_func


def new_choice_func(seed):
    random_state = np.random.RandomState(seed)

    def choice_func(choice_list):
        return random_state.choice(choice_list)

    return choice_func


"""
Helper functions from
https://stackoverflow.com/questions/192109/is-there-a-built-in-function-to-print-all-the-current-properties-and-values-of-a
to print all attributes of a class without me explicitly coding it out.
"""


def api(obj):
    return [name for name in dir(obj) if name[0] != "_"]


def attrs(obj):
    disallowed_properties = {
        name
        for name, value in getmembers(type(obj))
        if isinstance(value, (property, FunctionType))
    }
    return {
        name: getattr(obj, name)
        for name in api(obj)
        if name not in disallowed_properties and hasattr(obj, name)
    }


def base_model_copy(prior: BaseModel, values: dict):

    new_dict = prior.dict()
    new_dict.update(values)
    debug(values)
    replacement = type(prior)(**new_dict)
    return replacement


end_all(globals())



import abc
from mesa.time import BaseScheduler


class IEconomy(abc.ABC):
    @abc.abstractmethod
    def step(self):
        pass

    @abc.abstractmethod
    def start_step(self):
        pass

    @abc.abstractmethod
    def end_step(self):
        pass



# isort: on
import random

import numpy as np
from aiondao._imports import *

# from aiondao.agents.dummy import DummyAgent
from aiondao.agents.commons import (
    Common,
    CommonAgent,
    convert_80p_to_cliff_and_halflife,
    create_token_batches,
)
from aiondao.libs.states import CommonsConfig, SimState
from aiondao.utils.networking import bootstrap_network, get_participants
from devtools import debug
from loguru import logger
from mesa import Agent, Model
from mesa.datacollection import DataCollector
from mesa.space import MultiGrid
from mesa.time import RandomActivation, StagedActivation

"""
If you want to perform a parameter sweep, call batch_run.py instead of run.py.
For details see batch_run.py in the same directory as run.py.
"""

# Start of datacollector functions


def get_num_rich_agents(model):
    """return number of rich agents"""

    rich_agents = [a for a in model.schedule.agents if a.savings > model.rich_threshold]
    return len(rich_agents)


def get_num_poor_agents(model):
    """return number of poor agents"""

    poor_agents = [a for a in model.schedule.agents if a.loans > 10]
    return len(poor_agents)


def get_num_mid_agents(model):
    """return number of middle class agents"""

    mid_agents = [
        a
        for a in model.schedule.agents
        if a.loans < 10 and a.savings < model.rich_threshold
    ]
    return len(mid_agents)


def get_total_savings(model):
    """sum of all agents' savings"""

    agent_savings = [a.savings for a in model.schedule.agents]
    # return the sum of agents' savings
    return np.sum(agent_savings)


def get_total_wallets(model):
    """sum of amounts of all agents' wallets"""

    agent_wallets = [a.wallet for a in model.schedule.agents]
    # return the sum of all agents' wallets
    return np.sum(agent_wallets)


def get_total_money(model):
    # sum of all agents' wallets
    wallet_money = get_total_wallets(model)
    # sum of all agents' savings
    savings_money = get_total_savings(model)
    # return sum of agents' wallets and savings for total money
    return wallet_money + savings_money


def get_total_loans(model):
    # list of amounts of all agents' loans
    agent_loans = [a.loans for a in model.schedule.agents]
    # return sum of all agents' loans
    return np.sum(agent_loans)



from abc import ABC
import abc
from typing import List, Optional, Union
from aiondao.libs.states import SimState
from aiondao.interfaces.simulate import ISimBlock, ISimTransform, IPipeline


class IPolicyBlock(ISimBlock):
    # Figure out if there's some policies
    def transform(self, state_space: SimState) -> SimState:

        return state_space


class IPolicyTransform(ISimTransform):

    # Figure out if there's some policies
    def transform(self, state_space: SimState) -> SimState:

        return super().transform(state_space)



import random
from typing import Any, Callable, Dict, List, Optional
from aiondao._imports import *
from dataclasses_json import config


import networkx as nx

from aiondao.utils._utilz import (
    new_probability_func,
    new_exponential_func,
    new_gamma_func,
    new_random_number_func,
    new_choice_func,
)

from aiondao.agents.commons import (
    CommonAgent as CommonDAO,
    convert_80p_to_cliff_and_halflife,
)


class CommonsConfig(BaseModel):
    """
    There are just so many options that passing them via kwargs has become
    rather unwieldy. Also it would be useful to have some footnotes next to each
    parameter.

    Inject this paramter into class to control simulation parameters.
    """

    #
    hatch_tribute: float = 0.2
    exit_tribute: float = 0.35
    proposals: int = 2
    hatchers: int = 5
    kappa: int = 2
    vesting_80p_unlocked: int = 120
    days_to_80p_of_max_voting_weight: int = 10
    # maximum fraction of the funding pool that a proposal can ever request
    max_proposal_request: float = 0.2
    timesteps_days: int = 730
    random_seed: Optional[None] = Field(None)
    rand_seed: Optional[Callable] = None
    probability_func: Optional[Callable] = None
    exponential_func: Optional[Callable] = None
    gamma_func: Optional[Callable] = None
    random_number_func: Optional[Callable] = None
    choice_func: Optional[Callable] = None
    run_time: int = 20
    speculation_days: int = 120
    multiplier_new_participants: float = 0.2

    @root_validator
    def calculate_numerical_functions(cls, values):
        rand_seed = values.get("random_seed", None)
        probability_func = new_probability_func(rand_seed)
        exponential_func = new_exponential_func(rand_seed)
        gamma_func = new_gamma_func(rand_seed)
        random_number_func = new_random_number_func(rand_seed)
        choice_func = new_choice_func(rand_seed)

        values["probability_func"] = probability_func
        values["exponential_func"] = exponential_func
        values["gamma_func"] = gamma_func
        values["random_number_func"] = random_number_func
        values["choice_func"] = choice_func

        vest_unlocked_80 = values.get("vesting_80p_unlocked", 120)

        values["speculation_days"] = int(0.2 * vest_unlocked_80) + int(
            0.6 * vest_unlocked_80 * random_number_func()
        )
        values["multiplier_new_participants"] = 1 + int(9 * random_number_func())

        return values

    def alpha(self) -> float:
        """
        Converts days_to_80p_of_max_voting_weight to alpha.
        alpha = (1 - 0.8) ^ (1/t)
        """
        return 0.2 ** (1 / self.days_to_80p_of_max_voting_weight)

    def cliff_and_halflife(self) -> Tuple[float, float]:
        """
        This is just a wrapper around hatch.convert_80p_to_cliff_and_halflife().
        It is used here instead for logical consistency (having all derived
        configuration variables come from CommonsSimulationConfiguration), but
        defined in hatch for local consistency (having all hatch-related code in
        one place). Time will tell if this really matters.
        """
        return convert_80p_to_cliff_and_halflife(self.vesting_80p_unlocked)


class SimulationCalls:
    def __init__(self, state: "SimState") -> None:
        self.state = state

    def prob_fn(self, prob: float) -> float:
        return self.state.confs.probability_func(prob)  # type: ignore

    def exp_fn(self, loc: float, scale: float) -> float:
        return self.state.confs.exponential_func(loc=loc, scale=scale)  # type: ignore


class SimState(BorgModel):

    unique_id: int
    network: nx.DiGraph
    commons: CommonDAO
    funding_pool: float
    collateral_pool: float
    token_supply: float
    token_price: float
    policy_output: Optional[Dict[str, Any]] = {}
    sentiment: float = 0.75
    policy_log: List[Dict[str, Any]] = []
    timestep: Union[float, int] = 0
    testing_adder: int = 0
    confs: CommonsConfig = None
    conviction_proposals: List[Any] = []

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow

    def to_fns(self) -> SimulationCalls:
        return SimulationCalls(self)


class StepSpace(BorgModel):

    update_holdings_msg: bool = False
    crazy_id: int = 444
    # network: nx.DiGraph
    # commons: CommonDAO
    # funding_pool: float
    # collateral_pool: float
    # token_supply: float
    # token_price: float
    # policy_output: Optional[Dict[str, Any]] = None
    # sentiment: float = 0.75
    # policy_log: List[Dict[str, Any]] = []
    # timestep: Union[float, int] = 0
    # testing_adder: int = 0
    # confs: CommonsConfig = None

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow

    # def to_fns(self) -> SimulationCalls:
    #     return SimulationCalls(self)


"""

initial_conditions = {
            "network": network,
            "commons": commons,
            "funding_pool": commons._funding_pool,
            "collateral_pool": commons._collateral_pool,
            "token_supply": commons._token_supply,
            "token_price": commons.token_price(),
            "policy_output": None,
            "sentiment": 0.75,
        }
"""



from typing import Tuple
from aiondao._imports import *
from aiondao.libs.abcurve import AugmentedBondingCurve
from collections import namedtuple
from aiondao.utils._utilz import attrs
from aiondao import config


def vesting_curve(day: int, cliff_days: int, halflife_days: float) -> float:
    """
    The vesting curve includes the flat cliff, and the halflife curve where tokens are gradually unlocked.
    It looks like _/--
    """
    return 1 - config.vesting_curve_halflife ** ((day - cliff_days) / halflife_days)


def convert_80p_to_cliff_and_halflife(
    days: int, v_ratio: int = 2
) -> Tuple[float, float]:
    """
    For user's convenience, we ask him after how many days he would like 80% of his tokens to be unlocked.
    This needs to be converted into a half life (unit days).
    2.321928094887362 is log(base0.5) 0.2, or log 0.2 / log 0.5.
    v_ratio is cliff / halflife, and its default is determined by Commons Stack
    """
    halflife_days = days / (config.log_base05_of_02 + v_ratio)
    cliff_days = v_ratio * halflife_days
    return cliff_days, halflife_days


def hatch_raise_split_pools(total_hatch_raise, hatch_tribute) -> Tuple[float, float]:
    """Splits the hatch raise between the funding / collateral pool based on the fraction."""
    funding_pool = hatch_tribute * total_hatch_raise
    collateral_pool = total_hatch_raise * (1 - hatch_tribute)
    return funding_pool, collateral_pool


VestingOptions = namedtuple("VestingOptions", "cliff_days halflife_days")


class TokenBatch:
    def __init__(self, vesting: float, nonvesting: float, vesting_options=None):
        self.vesting = vesting
        self.nonvesting = nonvesting
        self.vesting_spent = 0.0

        self.age_days = 0

        self.cliff_days = 0 if not vesting_options else vesting_options.cliff_days
        self.halflife_days = 0 if not vesting_options else vesting_options.halflife_days

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, attrs(self))

    @property
    def total(self):
        return (self.vesting - self.vesting_spent) + self.nonvesting

    def __bool__(self):
        if self.total > 0:
            return True
        return False

    def __add__(self, other):
        total_vesting = self.vesting + other.vesting
        total_nonvesting = self.nonvesting + other.nonvesting
        return total_vesting, total_nonvesting

    def __sub__(self, other):
        total_vesting = self.vesting - other.vesting
        total_nonvesting = self.nonvesting - other.nonvesting
        return total_vesting, total_nonvesting

    def update_age(self, iterations: int = 1):
        """
        Adds the number of iterations to TokenBatch.age_days
        """
        self.age_days += iterations
        return self.age_days

    def unlocked_fraction(self) -> float:
        """
        returns what fraction of the TokenBatch is unlocked to date
        """
        if self.cliff_days and self.halflife_days:
            u = vesting_curve(self.age_days, self.cliff_days, self.halflife_days)
            return u if u > 0 else 0
        else:
            return 1.0

    def spend(self, x: float):
        """
        checks if you can spend so many tokens, then decreases this TokenBatch
        instance's value accordingly
        """
        if x > self.spendable():
            raise Exception(
                "Not so many tokens are available for you to spend yet ({})".format(
                    self.age_days
                )
            )

        y = x - self.nonvesting
        if y > 0:
            self.vesting_spent += y
            self.nonvesting = 0.0
        else:
            self.nonvesting = abs(y)

        return self.vesting, self.vesting_spent, self.nonvesting

    def spendable(self) -> float:
        """
        spendable() = (self.unlocked_fraction * self.vesting - self.vesting_spent) + self.nonvesting
        Needed in case some Tokens were burnt before.
        """
        return (
            (self.unlocked_fraction() * self.vesting) - self.vesting_spent
        ) + self.nonvesting


def create_token_batches(
    hatcher_contributions: List[int],
    desired_token_price: float,
    cliff_days: float,
    halflife_days: float,
) -> Tuple[List[TokenBatch], float]:
    """
    hatcher_contributions: a list of hatcher contributions in DAI/ETH/whatever
    desired_token_price: used to determine the initial token supply
    vesting_80p_unlocked: vesting parameter - the number of days after which 80% of tokens will be unlocked, including the cliff period
    """
    total_hatch_raise = sum(hatcher_contributions)
    initial_token_supply = total_hatch_raise / desired_token_price

    # In the hatch, everyone buys in at the same time, with the same price. So just split the token supply amongst the hatchers proportionally to their contributions
    tokens_per_hatcher = [
        (x / total_hatch_raise) * initial_token_supply for x in hatcher_contributions
    ]

    token_batches = [
        TokenBatch(x, 0, vesting_options=VestingOptions(cliff_days, halflife_days))
        for x in tokens_per_hatcher
    ]
    return token_batches, initial_token_supply


@dataclass(config=autovalid)
class Common:
    hatch_tribute: float
    total_hatch_raise: float
    token_supply: float
    exit_tribute: float
    kappa: int

    def __post_init__(self):
        # a fledgling commons starts out in the hatching phase. After the hatch phase ends, money from new investors will only go into the collateral pool.
        # Essentials
        self.collateral_pool = (1 - self.hatch_tribute) * self.total_hatch_raise
        self.funding_pool = (
            self.hatch_tribute * self.total_hatch_raise
        )  # 0.35 * total_hatch_raise = 35%
        # hatch_tokens keeps track of the number of tokens that were created when hatching, so we can calculate the unlocking of those
        self.hatch_tokens = self.token_supply
        self.bonding_curve = AugmentedBondingCurve(
            init_reserve=self.collateral_pool,
            supply=self.token_supply,
            kappa=self.kappa,
        )

        # Options
        self.exit_tribute = self.exit_tribute


class Commons:
    def __init__(
        self,
        total_hatch_raise,
        token_supply,
        hatch_tribute=0.2,
        exit_tribute=0,
        kappa=2,
    ):
        # a fledgling commons starts out in the hatching phase. After the hatch phase ends, money from new investors will only go into the collateral pool.
        # Essentials
        self.hatch_tribute = hatch_tribute
        # (1-0.35) -> 0.65 * total_hatch_raise = 65% collateral, 35% funding
        self._collateral_pool = (1 - hatch_tribute) * total_hatch_raise
        self._funding_pool = (
            hatch_tribute * total_hatch_raise
        )  # 0.35 * total_hatch_raise = 35%
        self._token_supply = token_supply
        # hatch_tokens keeps track of the number of tokens that were created when hatching, so we can calculate the unlocking of those
        self._hatch_tokens = token_supply
        self.bonding_curve = AugmentedBondingCurve(
            self._collateral_pool, token_supply, kappa=kappa
        )

        # Options
        self.exit_tribute = exit_tribute

    def deposit(self, dai):
        """
        Deposit DAI after the hatch phase. This means all the incoming deposit goes to the collateral pool.
        """
        tokens, realized_price = self.bonding_curve.deposit(
            dai, self._collateral_pool, self._token_supply
        )
        self._token_supply += tokens
        self._collateral_pool += dai
        return tokens, realized_price

    def burn(self, tokens):
        """
        Burn tokens, with/without an exit tribute.
        """
        dai, realized_price = self.bonding_curve.burn(
            tokens, self._collateral_pool, self._token_supply
        )
        self._token_supply -= tokens
        self._collateral_pool -= dai
        money_returned = dai

        if self.exit_tribute:
            self._funding_pool += self.exit_tribute * dai
            money_returned = (1 - self.exit_tribute) * dai

        return money_returned, realized_price

    def dai_to_tokens(self, dai):
        """
        Given the size of the common's collateral pool, return how many tokens would x DAI buy you.
        """
        price = self.bonding_curve.get_token_price(self._collateral_pool)
        return dai / price

    def token_price(self):
        """
        Query the bonding curve for the current token price, given the size of the commons's collateral pool.
        """
        return self.bonding_curve.get_token_price(self._collateral_pool)

    def spend(self, amount):
        """
        Decreases the Common's funding_pool by amount.
        Raises an exception if this would make the funding pool negative.
        """
        if self._funding_pool - amount < 0:
            raise Exception(
                "{} funds requested but funding pool only has {}".format(
                    amount, self._funding_pool
                )
            )
        self._funding_pool -= amount
        return
